"""
        Action find package duplicate
"""

import collections
from lib.actions import Action
from .actionsearch import ActionSearch
from lib.alpm import AlpmBranch
from lib.exceptions import BranchError


class ActionDuplicate(Action):

    @staticmethod
    def usage():
        return "command : find package duplicates over repos"

    def test(self):
        super().test()
        if not self.params.branch:
            raise BranchError(f"Enter a valid branch {AlpmBranch.BRANCHES}")
        return True

    def run(self):
        """ find package duplicates over repos """
        super().run()
        self.params.options['--quiet'] = True   # for search
        for branchname in AlpmBranch.BRANCHES:
            if self.params.branch and self.params.branch != branchname:
                continue
            self.find_dupp(branchname)

    def find_dupp(self, branch_name: str):
        #self.params.branch = branch_name
        branch = AlpmBranch(branch_name)
        pkgs = sorted((p.name for p in branch.pkg_search('')))
        dupps = [p for p, count in collections.Counter(pkgs).items() if count > 1]
        search = ActionSearch(self.params)
        if dupps:
            print("")
            reg = ''
            for dupp in dupps:
                reg += f"^{dupp}$|"
            search.run(reg.rstrip('|'))



"""
    Action check deps
"""

import subprocess
from lib.actions import Action
from lib.alpm import AlpmBranch


class ActionCheck(Action):

    @staticmethod
    def usage():
        return "check : Check dependencies"

    def test(self):
        super().test()
        return True

    def run(self):
        """ Check dependencies """
        super().run()
        for branchname in AlpmBranch.BRANCHES:
            if self.params.branch and self.params.branch != branchname:
                continue
            self.msg(f"Check dependencies {AlpmBranch.color(branchname)}{branchname}{AlpmBranch.color('none')}")
            subprocess.call(['pacman', '-Dkk', '--config', f"{AlpmBranch.ROOT}/{branchname}/etc/pacman.conf"])



"""
        Action files
"""

from lib.actions import Action
from lib.alpm import AlpmBranch
from lib.exceptions import BranchError


class ActionFiles(Action):

    @staticmethod
    def usage():
        return "files regex --unstable : search in files path\n" + \
            "\tif absolute path can find in all branches"

    def test(self):
        super().test()
        if not self.params.values:
            raise ValueError(f"Nothing to find !")
        if len(self.params.values[0]) < 3:
            raise ValueError(f"Nothing to find !")
        if not self.params.branch and not self.params.values[0].startswith('/'):
            raise BranchError(f"Enter a valid branch {AlpmBranch.BRANCHES}")
        return True

    def run(self):
        """ search in packages files """
        super().run()
        to_find = self.params.values.pop(0)
        for branch_name in AlpmBranch.BRANCHES:
            if self.params.branch and self.params.branch != branch_name:
                continue
            branch = AlpmBranch(branch_name)
            self.msg(f"Branch: {branch.color(branch_name)}{branch_name}{branch.color('none')}")
            try:
                branch.find_files(to_find, not self.params.branch)
            except FileNotFoundError as error:
                if self.params.branch:
                    raise error
                else:
                    print('\tNo files database ...')



"""
    Action chech MA profiles
"""

import os
from pathlib import Path
import urllib
import lib
from lib.actions import Action
from lib.alpm import AlpmBranch


class ActionProfiles(Action):

    @staticmethod
    def usage():
        return "profiles : Check manjaro m-a iso profiles"

    def test(self):
        # not force one branch ...
        super().test()
        return True

    def list_desktops_url(self, desktop_filters):
        """
            return generator on all desktop
        """
        desks = "xfce", "kde", "gnome"
        for desk in desks:
            if desktop_filters and desk not in desktop_filters:
                continue
            yield "manjaro/" + desk
        desks = (
            "openbox-regular",
            "mate",
            "bspwm-mate",
            "openbox-minimal",
            "deepin",
            "bspwm",
            "cinnamon",
            "lxqt",
            "openbox-basic",
            "awesome",
            "budgie",
            "lxde",
            "i3",
            "webdad"
        )
        for desk in desks:
            if desktop_filters and desk not in desktop_filters:
                continue
            yield "community/" + desk

    def extract_pkg_from_profile(self, file_name: str):
        """
            parse desktop file for extract packages
        """
        result = []
        with open(file_name, 'rt') as f_pro:
            for line in f_pro:
                if '>i686' in line:
                    continue
                line = line.strip()
                if not line or line.startswith('#'):
                    continue
                result.append(line.split('#')[0].strip().split()[-1])
        return frozenset(sorted(result))

    def run(self):
        """
            check packages in iso profiles
        """
        super().run()
        # we can add profile desktop in parameter for test only some
        desktop_filters = self.params.values
        if desktop_filters:
            desktop_filters = set(x.lower() for x in desktop_filters)
            print("Download only:", desktop_filters)
        local_profil = self.params.option('profil', 'p') # not used, not load but use local directory

        # download MA profiles
        profiles = []
        for desk in self.list_desktops_url(desktop_filters):
            url = f"https://gitlab.manjaro.org/profiles-and-settings/iso-profiles/raw/manjaro-architect/{desk}/Packages-Desktop"
            print("url:", url)
            filename = f"{lib.TMP_DIR}/iso-profiles/{desk}/Packages-Desktop"
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            try:
                urllib.request.urlretrieve(url, filename)
            except urllib.error.HTTPError:
                pass

        # find all profiles
        for f_packages in Path(f"{lib.TMP_DIR}/iso-profiles").glob('*/*/Packages-Desktop'):
            profiles.append(f_packages)

        def check_branch(branch_name: str):
            """ test for one branch """
            print("")
            branch = AlpmBranch(branch_name)
            pkgs = sorted([p.name for p in branch.pkg_search('')])
            self.msg(f"Branch: {branch.color(branch_name)}{branch_name}{branch.color('none')}")
            for profile in profiles:
                pkgs_profile = self.extract_pkg_from_profile(profile)
                # return packages are not in pkgs(packages in prepos)
                only_in = pkgs_profile - set(pkgs)
                print("\n->", profile.parent.name.upper())
                if only_in:
                    self.msg(f"{len(only_in)}:", *only_in)

        for branchname in AlpmBranch.BRANCHES:
            if branchname == 'archlinux':
                continue
            if self.params.branch and self.params.branch != branchname:
                continue
            check_branch(branchname)



"""
    Action display help
"""

import lib
from lib.actions import Action, ActionPlugin
#from lib.alpm import AlpmBranch


class ActionUsage(Action):

    @staticmethod
    def usage():
        return "General Help"

    def run(self):
        """display general usage"""
        super().run()
        #sep = "------"
        print(f":: {lib.DESCRIPTION} {lib.version_str()}")
        print(f"run `{lib.name()} -h` for global help or `{lib.name()} \"action\" -h` for action help")
        txt = """
                versions [packager] --unstable --archlinux [-q] [--verbose | -v]: compare version over branches
                diff --testing --stable [--rm | --add] : compare two branches, can show only one column
                check : Check dependencies
                duplicates : find package duplicates over repos
                profiles : Check manjaro m-a profiles
                ------
                info pkgname --unstable [--quiet or -q] [--files [--filter value]] : pacman -Si one branch
                search regex --unstable [--desc] [--quiet or -q] [--packager] [--dep] : search in one branch, by default only in pkgname
                files regex --unstable : search in files path
                pactree : as original command, only --sync
                ------
                tree : local database tree informations
                gist : last boxit annonces
                list [--unstable] : packagers listing
                ------
                update [--force or -f] [--files] [--mirrors | -m] : update all branches
                rm : remove all local databases"""
        print(txt.replace('               ', ''))

        print()
        for a in lib.action.ACTIONS:
            aa = lib.action.ACTIONS[a]
            if isinstance(aa, lib.actions.ActionPlugin):
                print(f" *{a} :", aa.usage())



"""
Application Exceptions
"""

class BranchNotFoundError(FileNotFoundError):
    """ Exception local database not set on disk """
    def __init__(self, message, branch):
        super().__init__(message)
        self.branch = branch
        self.message = message

class BranchError(ValueError):
    """ Exception bad branch """
    def __init__(self, message, branch=""):
        super().__init__(message)
        self.branch = branch
        self.message = message

class TemplateError(ValueError):
    """ Exception bad display template """
    def __init__(self, message, template=""):
        super().__init__(message)
        self.template = template
        self.message = message



"""
    Action versions - compare versions
"""

# TODO html output
# out_html = self.params.option('html')

import os
import shutil
import urllib
import gzip
import pyalpm
import lib
from lib.actions import Action
from lib.action.actionpactree import ActionPactree
from lib.alpm import AlpmBranch
from lib.exceptions import BranchError


class ActionVersions(Action):

    @staticmethod
    def usage():
        return """
            versions [packager] --unstable --archlinux [-q] [--verbose | -v]: compare version over branches
                packager : stefano@manjaro
                -q : only package name
                --verbose | -v : show all differences
                --aur : only if archlinux selected
                --pkg [pkgname] : test package dependencies
            """

    def test(self):
        super().test()
        if not self.params.branch:
            raise BranchError(f"Enter a valid branch {AlpmBranch.BRANCHES}")
        if not self.params.branch_two:
            raise BranchError(f"Enter a second valid branch {AlpmBranch.BRANCHES}")
        return True

    def run(self):
        """
        command : Maintener test versions over branch
        versions [user] [--aur]
        """
        super().run()
        branches = self.params.get_branches()
        use_aur = self.params.option('aur')
        verbose = self.params.option('verbose', 'v')
        quiet = self.params.option('quiet', 'q')
        packager = ''
        use_pkg = self.params.option('pkg', 'p')
        if use_pkg and isinstance(use_pkg, bool):
            try:
                use_pkg = self.params.values.pop(0).strip()
            except IndexError:
                use_pkg = ''
        else:
            try:
                packager = self.params.values.pop(0).strip()
            except IndexError:
                packager = ''
                if use_aur:
                    print('No AUR if not packager !')
                use_aur = False     # for big aur request

        branch = AlpmBranch(branches[1])
        branch_two = AlpmBranch(branches[0])

        print(f":: Compare versions {branch.color(branch.branch)}{branch.branch}{branch.color('none')} / {branch_two.color(branch_two.branch)}{branch_two.branch}{branch.color('none')}")
        pkgs = frozenset(p for p in branch.pkg_search('') if packager in p.packager)

        if use_pkg:
            try:
                old_branch = self.params.branch
                self.params.branch = 'unstable'
                ret = ActionPactree(self.params).run(use_pkg)
            finally:
                self.params.branch = old_branch
            if ret:
                ret = set(ret) & set(p.name for p in pkgs)
                pkgs = frozenset(p for p in pkgs if p.name in ret)
                print(f":: Package {branch.color(branch.branch)}{use_pkg}{branch.color('none')}")
            else:
                raise UserWarning(f"Nothing found ! {use_pkg}")
        aurs = {}

        if use_aur and branch_two.branch == 'archlinux':
            # download list all packages in AUR
            if not os.path.exists(f"{lib.TMP_DIR}/aur.packages"):
                if not os.path.exists(lib.TMP_DIR):
                    os.mkdir(lib.TMP_DIR)
                with urllib.request.urlopen("https://aur.archlinux.org/packages.gz") as f_url:
                    with open(f"{lib.TMP_DIR}/packages.gz", 'wb') as p_archive:
                        p_archive.write(f_url.read())
                with gzip.open(f"{lib.TMP_DIR}/packages.gz", 'rb') as z_file, open(f"{lib.TMP_DIR}/aur.packages", 'wb') as t_file:
                    shutil.copyfileobj(z_file, t_file, 65536)
                os.remove(f"{lib.TMP_DIR}/packages.gz")
            if verbose:
                print(":: use AUR")

            aurs = self.load_aur_pkgs(f"{lib.TMP_DIR}/aur.packages", frozenset(p.name for p in pkgs))
            print(f"{len(aurs)} packages manjaro in aur")
            aurs = AlpmBranch.aur_infos(aurs, verbose)
            #print('aur_requests:', aurs)

        i = 0
        for i, pkg in enumerate(pkgs):
            pkg_arch = branch_two.pkg_get(pkg.name)
            in_aur = '    '
            if pkg_arch:
                version_two = pkg_arch.version
            else:
                version_two = aurs.get(pkg.name, pkg.version)
                if version_two != pkg.version:
                    in_aur = '(aur)'

            if version_two != pkg.version:
                cmp_versions = pyalpm.vercmp(pkg.version, version_two)
                if quiet:
                    if cmp_versions < 0:
                        print(pkg.name)
                        continue
                else:
                    if verbose or cmp_versions < 0:
                        if cmp_versions < 0:
                            version2 = f"{self.colorize(pkg.version, version_two)}  {in_aur}"
                        else:
                            version2 = f"{version_two}  {in_aur}"
                        print(f"{pkg.name:36} {pkg.version:28} {version2:38} {cmp_versions:>2}")

        if not quiet:
            print(f":: {i} packages")
        #if use_aur and verbose:
        #    print(f"\nAUR packages ({len(aurs)}):\n", *sorted(aurs), sep="  ")

    @staticmethod
    def colorize(ver1, ver2, sep='\33[31m', end='\33[0m') -> str:
        result = ''
        first = True
        for c in ver2:
            if ver1.startswith(f"{result}{c}"):
                result += c
            else:
                if first:
                    first = False
                    result += f"{sep}{c}"
                else:
                    result += c
        if first:
            result += f"{result}{sep}"  # for align columns
        return f"{result}{end}"

    @staticmethod
    def load_aur_pkgs(file_name, include_pkgs) -> list:
        """ find if packages are in file 'aur list packages' """
        result = []
        with open(file_name, 'rt') as t_file:
            for line in t_file:
                line = line.rstrip('\n')
                if line in include_pkgs:
                    result.append(line)
            return result



"""
    parse console arguments
"""

import re
import lib
from lib.alpm import AlpmBranch
from lib.exceptions import BranchError

class ConsoleParser:
    """Arguments pass by console"""

    def __init__(self, args):
        self._args = args[1:]
        if self._args and self._args[0] == '-v':
            print(lib.DESCRIPTION, ":", lib.version_str())
            exit(0)
        self.expend()
        self.parse_branches()
        self.action = 'update'
        self.branch = ''
        self.branch_two = ''
        self.options = {}
        self.values = []
        try:
            self.action = self._args.pop(0).lower()
        except IndexError:
            self.action = ''

        self.read_branch()
        self.parse()
        #self.show()

    def show(self):
        """for debug"""
        print("args", self._args)
        print("action", self.action)
        print("branch", self.branch)
        print("branch 2", self.branch_two)
        print("options", self.options)
        print("params", self.values)
        #exit(0)

    def expend(self):
        """ split short option to options """
        # convert -suq -> -s -u -q
        regex = re.compile(r'^-[a-z]{2,}')
        for i, arg in enumerate(self._args):
            if not regex.search(arg):
                continue
            self._args[i] = arg[0:2]
            long_arg = arg[2:].strip()
            for c_arg in long_arg:
                self._args.insert(i+1, f"-{c_arg}")

    def parse_branches(self):
        """convert short branch"""
        # switch = {'-a':'--archlinux', '-u':'--unstable', '-t':'--testing', '-g':'--stable-staging', '-s':'--stable'}
        switch = {'-a': '--archlinux', '-u': '--unstable', '-t': '--testing', '-s': '--stable'}
        for i, arg in enumerate(self._args):
            self._args[i] = switch.get(arg, arg)
        for i, arg in enumerate(self._args):
            if arg.lower() in switch.values():
                self._args[i] = arg.lower()

    def read_branch(self):
        """find branch in console argument"""
        self.branch = None
        for arg in self._args:
            if arg[2:] in AlpmBranch.BRANCHES:
                if not self.branch:
                    self.branch = arg[2:]
                else:
                    if self.branch != arg[2:]:
                        self.branch_two = arg[2:]
        if self.branch:
            self._args.remove(f"--{self.branch}")
        if self.branch_two:
            self._args.remove(f"--{self.branch_two}")

    def get_branches(self) -> str:
        """find branches in console argument and order"""
        if not self.branch_two:
            raise BranchError(f"set second branch !")
        if AlpmBranch.BRANCHES.index(self.branch) < AlpmBranch.BRANCHES.index(self.branch_two):
            return (self.branch, self.branch_two)
        else:
            return (self.branch_two, self.branch)

    def parse(self) -> None:
        """set params in values et options with arg"""
        # get values
        for i, arg in enumerate(self._args):
            if arg.startswith('-'):
                break
            else:
                self.values.append(arg)
                self._args[i] = '.'
        # get options
        for i, arg in enumerate(self._args):
            if arg == '.':
                continue
            if arg.startswith('-'):
                # if "-d5"
                if arg[2:] and not arg[2:].isalpha():
                    arg_data = arg[2:]
                    arg = arg[0:2]
                else:
                    arg_data = (self._args[i+1:i+2] or (True,))[0]
                    # test next param if is a value or option
                    if not isinstance(arg_data, bool):
                        if arg_data.startswith('-'):
                            arg_data = True
                        else:
                            self._args[i+1] = '.'
                self._args[i] = '.'
                self.options[arg] = arg_data

    def option(self, key: str, short_key: str = ''):
        """return option"""
        result = self.options.get(f"--{key}", False)
        if not result:
            result = self.options.get(f"-{short_key}", False)
        return result




"""
starter module
"""

import re
from pathlib import Path

__all__ = (
    '__version__',
    'version_str',
    'name',
    'PKGNAME',
    'DESCRIPTION',
    'TMP_DIR',
    'read_config_file',
    'read_config_root'
)

__version__ = (0, 9, 10, '')

def name():
    """ Provides a user-friendly name of the project """
    return "mbn"
    #return __name__

def version_str(version=__version__):
    """ Provides the version string for this project """
    result = '.'.join(map(str, version[:3]))
    if len(version) > 3:
        result += '-' + '-'.join(version[3:])
    return result

DESCRIPTION = "mbn - manjaro Multi-Branch Navigator"
TMP_DIR = f"/tmp/{name()}-multi-branches"
PKGNAME = "manjaro-check-repos"

def read_config_file(key: str, default: str = '') -> str:
    """where is installed pacman local databases"""
    result = default
    regex = re.compile(r'^' + key + '( ?)=(?P<path>[^#]+)')
    with open(f"/etc/{PKGNAME}/config.conf", 'r') as conf:
        for line in conf:
            if line.startswith("#"):
                continue
            matchs = regex.match(line)
            if not matchs:
                continue
            path = matchs.group('path')
            if path:
                result = path.strip().rstrip('/')
                break
    return result

def read_config_root() -> str:
    """where is installed pacman local databases"""
    path_root = read_config_file('RootDir', '/var/cache/manjaro-repos')
    path_root = Path(path_root)
    if not path_root.is_absolute():
        path_root = path_root.expanduser()
    try:
        path_root.mkdir(parents=True, exist_ok=True)
    except PermissionError:
        pass
    return str(path_root)



"""
    Action info
"""


import os
import urllib
from datetime import datetime
from urllib.parse import urlparse
import json
import lib
from lib.actions import Action
from .actionsearch import ActionSearch
from lib.alpm import AlpmBranch
from pycman import pkginfo


class ActionInfo(Action):

    @staticmethod
    def usage():
        return """
            info pkgname --unstable [--quiet or -q] [--files [--filter value]] : pacman -Si one branch
            info pkgname [--quiet or -q] : pacman -Si for all branches
            """

    def test(self):
        super().test()
        if not self.params.values:
            raise ValueError(f"No package to find !")
        return True

    def run(self):
        """ pacman -Qi in one or all branches """
        super().run()
        quiet = self.params.option('quiet', 'q')
        to_find = self.params.values.pop(0).lower()
        nb = [0]

        def get_info(branchname: str, pkgname: str, count: list):
            branch = AlpmBranch(branchname)
            pkg = branch.pkg_get(pkgname)
            if pkg:
                count[0] += 1
                txt = 'Branch'
                print(f"{txt:15}: {branch.color(branchname)}{branchname}{branch.color('none')}")
                if quiet:
                    print(f"{'Name':15}: {pkg.name}")
                    print(f"{'Version':15}: {pkg.version}")
                    print(f"{'Repository':15}: {pkg.db.name}")
                    print(f"{'Build Date ':15}: {datetime.fromtimestamp(pkg.builddate).strftime('%a %d %b %Y %X %Z')}")
                    print(f"{'Packager':15}: {pkg.packager}")
                else:
                    pkginfo.display_pkginfo(pkg, style='sync')
                return pkg
            return None

        if not self.params.branch:
            for br in AlpmBranch.BRANCHES:
                pkg = get_info(br, to_find, nb)
            AlpmBranch.aur_query(to_find)
        else:
            pkg = get_info(self.params.branch, to_find, nb)
            if self.params.branch != 'archlinux':
                self.show_archlinux(pkg)
            # display files
            if pkg and not quiet and self.params.option('files'):
                use_filter = self.params.option('filter')
                pkgfiles = AlpmBranch.read_files(self.params.branch, pkg)
                if use_filter and not isinstance(use_filter, bool):
                    pkgfiles = [p for p in pkgfiles if use_filter in p]
                if pkgfiles:
                    for pkgfile in pkgfiles:
                        print(pkgfile)
                    print("")

        if pkg and not quiet:
            if 'manjaro' in pkg.packager:
                url = f"https://gitlab.manjaro.org/packages/{pkg.db.name}/{pkg.name}"
                req = urllib.request.urlopen(url, timeout=5)
                # req.getcode() == always 200 !
                if req.geturl() == url:
                    print(f"{'Gitlab manjaro':15}: {url}")
            # version from last git tag
            if not self.params.branch:
                self.params.branch = 'unstable'
            self.show_gitlab(pkg, self.params.branch)
            self.show_github(pkg, self.params.branch)
            self.show_python(pkg, self.params.branch)
            self.show_snap(pkg, self.params.branch)
            self.show_flatpack(pkg, self.params.branch)

        if nb[0] < 1:
            # package not found ?
            if not self.params.branch:
                self.params.branch = 'unstable'
            self.params.options['--quiet'] = True
            ActionSearch(self.params).run(f"{to_find}.*")
            raise UserWarning(f"Nothing found !")


    def show_archlinux(self, pkg: object) ->bool:
        """ archlinux version """
        #can also use https://www.archlinux.org/packages/{pkg.db}/x86_64/{pkg.name}/json/
        if not pkg:
            return False
        branch = AlpmBranch('archlinux')
        pkga = branch.pkg_get(pkg.name)
        if pkga and pkg.version != pkga.version:
            print(f"{'Archlinux':15}: {branch.color(branch.branch)}{pkga.version}{branch.color('none')}")
            return True


    def show_python(self, pkg: object, branchname: str = 'unstable') ->bool:
        """ version from pypi.org """
        if not pkg.name.startswith('python-'):
            return
        if not branchname:
            branchname = 'unstable'
        try:
            with urllib.request.urlopen(f"https://pypi.org/pypi/{pkg.name[7:]}/json", timeout=7) as f_url:
                req = f_url.read()
                datas = json.loads(req)
                if datas:
                    print(f"\n{'Pypi.org':15}: {AlpmBranch.color(branchname)}{datas['info']['version']}{AlpmBranch.color('none')}")
                    return True
        except urllib.error.HTTPError:
            return False

    def show_gitlab(self, pkg: object, branchname: str = 'unstable') -> bool:
        """Display gitlab informations"""
        result = False
        if not 'gitlab' in pkg.url:
            return False
        if not branchname:
            branchname = 'unstable'
        id_gitlab = None
        namespace = urlparse(pkg.url).path.strip('/')
        if '/' in namespace:
            project = namespace.split('/')[-1]
        else:
            project = namespace

        value = lib.read_config_file(pkg.name, '-1').lower()
        if value != '-1':
            id_gitlab = value
        else:
            url = f"https://{urlparse(pkg.url).hostname}/api/v4/projects?scope=project&search={project}"
            try:
                with urllib.request.urlopen(url, timeout=8) as f_url:
                    req = f_url.read()
                    items = json.loads(req)
            except urllib.error.HTTPError:
                return False
            for item in items:
                if item['path_with_namespace'] == namespace:
                    id_gitlab = item['id']
                    #print('id_gitlab:', id_gitlab)
                    break

        if not id_gitlab:
            return False

        print(f"\n{AlpmBranch.color(branchname)}::{AlpmBranch.color('none')} from Gitlab")
        try:
            with urllib.request.urlopen(f"http://{urlparse(pkg.url).hostname}/api/v4/projects/{id_gitlab}/releases", timeout=7) as f_url:
                req = f_url.read()
                items = json.loads(req)
                for item in items:
                    print(f"{'last release':15}: {AlpmBranch.color(branchname)}{item['tag_name']}{AlpmBranch.color('none')} {item['name']}")
                    result = True
                    break
        except urllib.error.HTTPError:
            pass

        try:
            with urllib.request.urlopen(f"http://{urlparse(pkg.url).hostname}/api/v4/projects/{id_gitlab}/repository/tags", timeout=7) as f_url:
                req = f_url.read()
                items = json.loads(req)
                for item in items[:3]:
                    print(f"{'last tag':15}: {AlpmBranch.color(branchname)}{item['name']}{AlpmBranch.color('none')}")
                    result = True
        except urllib.error.HTTPError:
            pass

        try:
            with urllib.request.urlopen(f"http://{urlparse(pkg.url).hostname}/api/v4/projects/{id_gitlab}/repository/commits", timeout=7) as f_url:
                req = f_url.read()
                items = json.loads(req)
                for item in items[:1]:
                    print(f"{'last commit':15}: {AlpmBranch.color(branchname)}{item['committed_date'][:10]}{AlpmBranch.color('none')}")
                    result = True
        except urllib.error.HTTPError:
            pass

        return result

    def show_github(self, pkg: object, branchname: str = 'unstable') ->bool:
        # tag_name
        result = False
        if not 'github.com' in pkg.url:
            return result
        if not branchname:
            branchname = 'unstable'
        headers = {
            'Accept': 'application/vnd.github.quicksilver-preview+json',
            'User-Agent': 'manjaro Multi Branch Navigator',
            'Authorization': f"token {os.environ.get('GITHUB_TOKEN')}"
        }
        items = None
        namespace = urlparse(pkg.url).path.strip('/')

        # get releases
        try:
            req = urllib.request.Request(f"https://api.github.com/repos/{namespace}/releases/latest", headers=headers)
            with urllib.request.urlopen(req, timeout=7) as f_url:
                req = f_url.read()
                items = json.loads(req)
                if items and items['tag_name']:
                    print(f"\n{AlpmBranch.color(branchname)}::{AlpmBranch.color('none')} from Github")
                    print(f"{'last release':15}: {AlpmBranch.color(branchname)}{items['tag_name']}{AlpmBranch.color('none')} {items['name']}")
                    result = True
        except urllib.error.HTTPError:
            pass

        # get tags
        try:
            req = urllib.request.Request(f"https://api.github.com/repos/{namespace}/git/refs/tags", headers=headers)
            with urllib.request.urlopen(req, timeout=7) as f_url:
                req = f_url.read()
                tags = json.loads(req)
                if not result and tags:
                    print(f"\n{AlpmBranch.color(branchname)}::{AlpmBranch.color('none')} from Github")
                for atag in reversed(tags[-3:]):
                    tag = atag['ref'].split('/')[-1]
                    if items and tag == items['tag_name']:
                        break
                    print(f"{'last tag':15}: {AlpmBranch.color(branchname)}{tag}{AlpmBranch.color('none')}")
                    result = True
        except urllib.error.HTTPError:
            pass

        #last commit
        try:
            req = urllib.request.Request(f"https://api.github.com/repos/{namespace}/commits/master", headers=headers)
            with urllib.request.urlopen(req, timeout=7) as f_url:
                req = f_url.read()
                lcommit = json.loads(req)
                if not result and lcommit:
                    print(f"\n{AlpmBranch.color(branchname)}::{AlpmBranch.color('none')} from Github")
                ldate = lcommit['commit']['author']['date'][:10]
                print(f"{'last commit':15}: {AlpmBranch.color(branchname)}{ldate}{AlpmBranch.color('none')}")
                result = True
        except urllib.error.HTTPError:
            pass
        return result


    def show_snap(self, pkg: object, branchname: str = 'unstable') ->bool:
        url = f"https://search.apps.ubuntu.com/api/v1/search?q={pkg.name}"
        try:
            req = urllib.request.Request(url)
            with urllib.request.urlopen(req, timeout=7) as f_url:
                req = f_url.read()
                items = json.loads(req)['_embedded']
                for item in items['clickindex:package']:
                    print(f"\n{AlpmBranch.color(branchname)}::{AlpmBranch.color('none')} from Snap")
                    print(f"{'release':15}: {AlpmBranch.color(branchname)}{item['version']} {AlpmBranch.color('none')} {item['title']} {item['date_published'][:10]}")
                    #return True
        except urllib.error.HTTPError:
            return False

    def show_flatpack(self, pkg: object, branchname: str = 'unstable') ->bool:
        try:
            req = urllib.request.Request('https://flathub.org/api/v1/apps/')
            with urllib.request.urlopen(req, timeout=7) as f_url:
                req = f_url.read()
                for item in json.loads(req):
                    if f".{pkg.name}" in item['flatpakAppId'].lower():
                        print(f"\n{AlpmBranch.color(branchname)}::{AlpmBranch.color('none')} from FlatHub")
                        print(f"{'release':15}: {AlpmBranch.color(branchname)}{item['currentReleaseVersion']} {AlpmBranch.color('none')} {item['name']} {item['currentReleaseDate'][:10]}")
                        #return True
        except urllib.error.HTTPError:
            return False

    '''

    def show_appimage(self, pkg: object, branchname: str = 'unstable') ->bool:
        try:
            req = urllib.request.Request('https://appimage.github.io/feed.json')
            with urllib.request.urlopen(req, timeout=7) as f_url:
                req = f_url.read()
                for item in json.loads(req)['items']:
                    if f"{pkg.name}" in item['name'].lower():
                        print(f"\n{AlpmBranch.color(branchname)}::{AlpmBranch.color('none')} as Appimage")
                        # no version information !!!
                        #return True
        except urllib.error.HTTPError:
            return False
    '''



#!/usr/bin/python
"""
    Application
"""

import sys
import subprocess
from lib.exceptions import BranchError
from lib.alpm import AlpmBranch
from lib import TMP_DIR
from mainapp import MainApp
from lib.action.actionbranch import ActionBranch

def error(code, err):
    """ display Exceptions """
    ico = {1: "🤥", 10: "🌿"}
    if code < 10:
        msg = str(err)
    else:
        msg = f"\33[31mError:{AlpmBranch.color('none')} {type(err).__name__}: {str(err)}"
    print("\n", msg, f"{ico.get(code,'🤕')}\n", file=sys.stderr)
    exit(code)

if __name__ == '__main__':
    DEBUG = False
    if DEBUG or '--DEBUG' in sys.argv:
        MainApp(sys.argv).run()
    else:
        try:
            MainApp(sys.argv).run()
        except UserWarning as err:
            error(1, err)
        except BranchError as err:
            print(ActionBranch.usage())
            error(10, err)
        except Exception as err:
            error(99, err)
        finally:
            subprocess.run(['rm', '-rf', TMP_DIR])
    print("")



"""
    Application Actions
"""
import sys
import os
import subprocess
import json
from lib.alpm import AlpmBranch

class Action():
    """ Général action """

    def __init__(self, params: object):
        """ class initialisation """
        self.params = params
        self.key = params.action
        self.error = ''

    @staticmethod
    def usage() -> str:
        """ return action help """
        return "Not implémented"

    def test(self) -> bool:
        """
        test parameters
        set self.error and/or create exception
        """
        if self.params.option("help", "h"):
            print(f"\n{self.usage().replace('           ', '')}\n")
            exit(0)
        return True

    def run(self):
        """ execute action """
        pass

    @property
    def color(self) -> str:
        """return color branch"""
        if self.params.branch:
            return AlpmBranch.color(self.params.branch)
        else:
            return AlpmBranch.color('none')

    def msg(self, *args):
        """format message with curent branch color"""
        print(f"{self.color}::{AlpmBranch.color('none')}", *args)


class ActionPlugin():
    path = "~/.config/mbn"
    def __init__(self, key: str = None):
        self.key = key
        self.params = {}
        self.usage()    # test if good plugin

    def test(self) -> bool:
        return True

    def usage(self):
        os.environ['PACMAN_ROOT'] = AlpmBranch.ROOT
        stdout, _ = subprocess.Popen(f"{self.path}/mbn-{self.key} get-plugin-metadata", shell=True, text=True, stdout=subprocess.PIPE).communicate()
        j = json.loads(stdout)
        return j["ShortDescription"]

    def run(self):
        os.environ['PACMAN_ROOT'] = AlpmBranch.ROOT
        if self.params.branch:
            os.environ['PACMAN_BRANCH'] = self.params.branch
        subprocess.run(f"{self.path}/mbn-{self.key} {' '.join(sys.argv)}", shell=True, text=True)

class SwitchClass():
    """ switch/case Class """

    def __init__(self, params: dict = None):
        """ class initialisation """
        self.classes = {}
        if params and isinstance(params, dict):
            self.classes = params

    def provide(self, key: str, class_type):
        """ add one action """
        if key:
            self.classes[key] = class_type
        return self

    def get(self, key: str, default=None):
        """ find class action """
        return self.classes.get(key, default)

    def factory(self, key: str, params: dict):
        """ create action object """
        action_type = self.get(key, self.get('usage'))
        if isinstance(action_type, ActionPlugin):
            action_type.params = params
            return action_type
        return action_type(params)



"""
    application actions
"""
from pathlib import Path
import json

from ..actions import ActionPlugin
from .actionusage import ActionUsage
from .actionbranch import ActionBranch
from .actionrm import ActionRm
from .actionlist import ActionList
from .actiontree import ActionTree
from .actionsearch import ActionSearch
from .actioninfo import ActionInfo
from .actiondiff import ActionDiff
from .actioncheck import ActionCheck
from .actionduplicate import ActionDuplicate
from .actiongist import ActionGist
from .actionpactree import ActionPactree
from .actionfiles import ActionFiles
from .actionupdate import ActionUpdate
from .actionprofiles import ActionProfiles
from .actionversions import ActionVersions


ACTIONS = {
    "usage": ActionUsage,
    "branch": ActionBranch,
    "list": ActionList,
    "update" : ActionUpdate,
    "rm": ActionRm,
    "tree": ActionTree,
    "search": ActionSearch,
    "info": ActionInfo,
    "versions": ActionVersions,
    "diff": ActionDiff,
    "check": ActionCheck,
    "duplicates": ActionDuplicate,
    "gist": ActionGist,
    "pactree": ActionPactree,
    "files": ActionFiles,
    "profiles": ActionProfiles,
}

for p in Path('~/.config/mbn/').expanduser().glob('mbn-*'):
    if not p.name[4:]:
        continue
    try:
        ACTIONS[p.name[4:]] = ActionPlugin(p.name[4:])
    except json.decoder.JSONDecodeError:
        print(f" ! plugin error {p.name}")
        continue



"""
        Action remove
"""
from lib.actions import Action


class ActionBranch(Action):

    @staticmethod
    def usage():
        return """
            branch:
                --archlinux or -a
                --unstable  or -u
                --testing  or -t
                --stable  or -s

            -ts == --testing --stable
            """

    def test(self):
        super().test()
        return False

    def run(self):
        """ nothing """
        super().run()



"""
        Action gist
"""

import urllib
import json
from lib.actions import Action
from lib.alpm import AlpmBranch


class ActionGist(Action):

    @staticmethod
    def usage():
        return "gist : last boxit annonces"

    def run(self):
        """ List annonces """
        super().run()
        url = "https://api.github.com/users/philmmanjaro/gists"
        try:
            with urllib.request.urlopen(url) as f_url:
                req = f_url.read().decode('utf-8')
        except urllib.error.HTTPError:
            return
        gists = json.loads(req)
        print("")
        filters = ['stable', 'testing', 'unstable']
        for filt in filters:
            print(f"{AlpmBranch.color(filt)}::{AlpmBranch.color('none')} {filt}")
            for gist in gists:
                desc = gist['description']
                if desc.lower().find(' update') < 1:
                    continue
                if desc.lower().find(f"[{filt}") < 0:
                    continue
                print(desc)
                for f_url in gist['files']:
                    url_file = gist['files'][f_url]['raw_url']
                    if url_file:
                        print(f"  {AlpmBranch.color('i')}{url_file}{AlpmBranch.color('none')}")
                    break
                break
        '''
        import xml.dom.minidom
        print("")
        for filter in filters:
            print(f"\n{AlpmBranch.color(filter)}::{AlpmBranch.color('none')} {filter}")
            url = "https://forum.manjaro.org/c/announcements/stable-updates.rss"
            with urllib.request.urlopen(url) as f:
                req = f.read().decode()
            dom = xml.dom.minidom.parseString(req)

            def getText(nodelist):
                rc = []
                for node in nodelist:
                    if node.nodeType == node.TEXT_NODE:
                        rc.append(node.data)
                return ''.join(rc)

            titles = dom.getElementsByTagName("title")
            for title in titles:
                desc = getText(title.childNodes)
                if desc.startswith('['):
                    print(desc)
                    break
            first = True
            titles = dom.getElementsByTagName("link")
            for title in titles:
                url_file = getText(title.childNodes)
                if not first:
                    print(f"  {AlpmBranch.color('i')}{url_file}{AlpmBranch.color('none')}")
                    break
                first = False
        '''
